r"""
This file is part of eduworld package.

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import tkinter as tk


# pylint: disable=too-many-ancestors
class Canvas(tk.Canvas):
    """Class for drawing the stuff on tk.canvas"""

    def __init__(self, parent):
        super().__init__(parent, width=320, height=240, background="gray95")
        self.root = parent
        self._draw_delay = 0.7
        self.grid(column=0, row=0, sticky="NEWS")
        self.bind("<Configure>", lambda _: self.redraw(True))

    def set_draw_delay(self, delay: float) -> None:
        """Sets the delay (like 0.7 or 0.05 of a second) between redraws"""
        self._draw_delay = min(max(0, delay), 1)

    def redraw(self, immediate=False) -> None:
        """Override this method in a derived class to draw something
        import time
        # then use following code in the override to draw with delay
        # --
        if not immediate:
            time.sleep(self.draw_delay)
        # draw something
        self.root.update()
        # ---
        """



r"""
This package provides the following functions defined in __all__ for writing
simple procedural style programs with AlgoWorld Robots

This file is part of eduworld package

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import sys
import tkinter as tk
import random as r

# from random import randint as rnd
from eduworld import Application, PlotterCanvas as Canvas


# pylint: disable=too-many-arguments
class _AppWithState:
    """Store App and it's state"""

    def __init__(self):
        self.cell_size: int = 35
        self._initial_cell_size: int = 35
        self.app: Application = Application()
        self.root = self.app.root
        self.canvas: Canvas = Canvas(self.root)
        self.initialized: bool = False
        self._colors = self._setup_colors()
        self.app.set_canvas(self.canvas)

    def setup(
        self,
        x: float = 0,
        y: float = 0,
        cell_size: int = 35,
        delay: float = 0.7,
        interactive: bool = False,
    ):
        """Setup the app"""
        if cell_size is not None:
            self.cell_size = cell_size
            self._initial_cell_size = cell_size
            self.canvas.set_cell_size(cell_size, False)
        self.canvas.set_draw_delay(delay)
        self.canvas.pen_set_pos(x, y)
        self._bind_movement_keys()
        if interactive:
            self._bind_paint_keys()
            self.canvas.set_draw_delay(0)

    def shutdown(self, keep_window: bool = False):
        """Kill the App"""
        self.canvas.set_draw_delay(0)
        self.app.shutdown(keep_window)

    def run(self):
        """Run the App"""
        self.app.run()

    def _setup_colors(self):
        return [
            "black",
            "gray",
            "red",
            "green",
            "blue",
            "dodgerblue",
            "cyan",
            "firebrick",
            "magenta",
            "orange",
            "pink",
        ]

    def _rand_color(self):
        i = r.randint(0, len(self._colors) - 1)
        self.canvas.pen_color(self._colors[i])

    def _incr_cell_size(self):
        self.cell_size += 2
        self.canvas.set_cell_size(self.cell_size)

    def _decr_cell_size(self):
        self.cell_size -= 2
        self.canvas.set_cell_size(self.cell_size)

    def _reset_pen(self):
        self.cell_size = self._initial_cell_size
        self.canvas.pen_set_pos(0, 0)
        self.canvas.set_cell_size(self._initial_cell_size)

    def _bind_movement_keys(self):
        self.root.bind("<KP_0>", lambda _: self._reset_pen())
        self.root.bind("<KP_7>", lambda _: self.canvas.pen_move_by(-1, 1))
        self.root.bind("<KP_9>", lambda _: self.canvas.pen_move_by(1, 1))
        self.root.bind("<KP_1>", lambda _: self.canvas.pen_move_by(-1, -1))
        self.root.bind("<KP_3>", lambda _: self.canvas.pen_move_by(1, -1))
        self.root.bind("<KP_2>", lambda _: self.canvas.pen_move_by(0, -1))
        self.root.bind("<KP_8>", lambda _: self.canvas.pen_move_by(0, 1))
        self.root.bind("<KP_4>", lambda _: self.canvas.pen_move_by(-1, 0))
        self.root.bind("<KP_6>", lambda _: self.canvas.pen_move_by(1, 0))
        self.root.bind("<KP_Add>", lambda _: self._incr_cell_size())
        self.root.bind("<KP_Subtract>", lambda _: self._decr_cell_size())

    def _bind_paint_keys(self):
        self.root.bind("<W>", lambda _: self.canvas.pen_width(r.randint(1, 15)))
        self.root.bind("<w>", lambda _: self.canvas.pen_width(r.randint(1, 15)))
        self.root.bind("<E>", lambda _: self._rand_color())
        self.root.bind("<e>", lambda _: self._rand_color())
        self.root.bind("<R>", lambda _: self.canvas.pen_raise())
        self.root.bind("<r>", lambda _: self.canvas.pen_raise())
        self.root.bind("<F>", lambda _: self.canvas.pen_lower())
        self.root.bind("<f>", lambda _: self.canvas.pen_lower())
        self.root.bind("<C>", lambda _: self.canvas.clear())
        self.root.bind("<c>", lambda _: self.canvas.clear())
        self.root.bind("<Z>", lambda _: self.canvas.undo_last_line())
        self.root.bind("<z>", lambda _: self.canvas.undo_last_line())


app: _AppWithState = _AppWithState()


def _hide_tk_err(f):
    def w(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except tk.TclError:
            sys.exit(1)

    return w


def setup(
    x: float = 0,
    y: float = 0,
    cell_size: int = None,
    delay: float = 0.7,
    interactive: bool = False,
) -> None:
    """Setup the plotter"""
    if not app.initialized:
        app.initialized = True
        app.setup(
            x=x,
            y=y,
            cell_size=cell_size,
            delay=delay,
            interactive=interactive,
        )
        app.run()


def shutdown(keep_window: bool = False) -> None:
    """Shut down the app"""
    app.shutdown(keep_window)


@_hide_tk_err
def pen_move_by(dx: float, dy: float) -> None:
    """Move pen by delta"""
    app.canvas.pen_move_by(dx, dy)


@_hide_tk_err
def pen_set_pos(x: float, y: float) -> None:
    """Move pen to apsolute x and y coordinates"""
    app.canvas.pen_set_pos(x, y)


def pen_raise() -> None:
    """Raise pen"""
    app.canvas.pen_raise()


def pen_lower() -> None:
    """Lower pen"""
    app.canvas.pen_lower()


def pen_color(color: str) -> None:
    """Set pen color"""
    app.canvas.pen_color(color)


def pen_width(width: int) -> None:
    """Set pen width"""
    app.canvas.pen_width(width)


__all__ = [
    # setup
    "setup",
    "shutdown",
    "pen_raise",
    "pen_lower",
    "pen_color",
    "pen_width",
    "pen_set_pos",
    "pen_move_by",
]



r"""
This file is part of eduworld package.

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import time

from ..canvas import Canvas


class Pen:
    """Simple data holder for Pen props"""

    def __init__(self):
        self.pos = [0, 0]
        self.color = "black"
        self.width = 6
        self.can_draw = False

    def move_by(self, dx: float, dy: float) -> None:
        """Move pen relatively to its position"""
        self.pos[0] += dx
        self.pos[1] += dy

    def set_pos(self, x: float, y: float) -> None:
        """Set absolute pen position"""
        self.pos[0] = x
        self.pos[1] = y

    def set_width(self, width: int) -> None:
        """Set pen width. Constrained between 4 .. 15"""
        self.width = int(max(4, min(width, 15)))


class Point:
    """Some arbitrary 2d point"""

    def __init__(self):
        self.x: int = 0
        self.y: int = 0


# pylint: disable=too-many-ancestors, too-many-instance-attributes
class PlotterCanvas(Canvas):
    """Class for drawing lines of many colors aka Plotter"""

    def __init__(self, parent):
        super().__init__(parent)
        self._unit_size = 35
        self._arrow_size = 16
        self._width = 0
        self._height = 0
        self._origin = Point()
        self._pen = Pen()
        self._lines = []

    def set_cell_size(self, cell_size: int, redraw: bool = True):
        """Set the unit cell size"""
        if cell_size is not None:
            self._unit_size = cell_size
        if redraw:
            self.redraw()

    def pen_raise(self):
        """Raise the pen. Plotter with pen raised can't draw lines"""
        self._pen.can_draw = False
        self.redraw()

    def pen_lower(self):
        """Lower the pen. Plotter with pen lowered can draw lines"""
        self._pen.can_draw = True
        self.redraw()

    def pen_color(self, color: str) -> None:
        """Set the pen color"""
        self._pen.color = color
        self.redraw()

    def pen_width(self, width: int) -> None:
        """Set the pen width"""
        self._pen.set_width(width)
        self.redraw()

    def pen_move_by(self, dx: float, dy: float) -> None:
        """Move pen by dx and dy"""
        p = self._pen
        last = p.pos.copy()
        p.move_by(dx, dy)
        if p.can_draw:
            coords = last + p.pos.copy()
            self._lines.append((coords, p.width, p.color))
        self.redraw()

    def pen_set_pos(self, x: float, y: float) -> None:
        """Set absolute pen position"""
        p = self._pen
        last = p.pos.copy()
        p.set_pos(x, y)
        if p.can_draw:
            coords = last + p.pos.copy()
            self._lines.append((coords, p.width, p.color))
        self.redraw()

    def undo_last_line(self):
        """Undo draw last line"""
        if len(self._lines) == 0:
            return;
        can_draw = self._pen.can_draw
        coords, width, color = self._lines.pop()
        self.pen_raise()
        self.pen_set_pos(coords[0], coords[1])
        self.pen_color(color)
        self.pen_width(width)
        if can_draw:
            self.pen_lower()
        self.redraw()

    def clear(self):
        """Remove all drawn lines"""
        self._lines.clear()
        self.redraw()

    def redraw(self, immediate=False) -> None:
        """Redraw everything"""
        if not immediate:
            time.sleep(self._draw_delay)
        self._height = self.winfo_height()
        self._width = self.winfo_width()
        self._origin.x = self._width / 2
        self._origin.y = self._height / 2
        self._delete_all()
        self._draw_grid()
        self._draw_lines()
        self._draw_pen()
        self._draw_origin_numbers()
        # self._draw_numbers()
        self.root.update()

    def _delete_all(self):
        self.delete("grid")
        self.delete("lines")
        self.delete("pen")
        self.delete("numbers")

    def _calc_true_origin(self):
        ox = self._origin.x
        oy = self._origin.y
        cx = -self._pen.pos[0] * self._unit_size
        cy = self._pen.pos[1] * self._unit_size
        return (ox + cx, oy + cy)

    def _draw_origin_numbers(self):
        ox, oy = self._calc_true_origin()
        hs = self._unit_size / 2
        s = self._unit_size
        tags = ("numbers",)
        font = "Mono 12 bold"
        fill = "gray"
        self.create_text(ox - hs, oy + hs, text="0", fill=fill, font=font, tags=tags)
        self.create_text(ox + s, oy + hs, text="1", fill=fill, font=font, tags=tags)
        self.create_text(ox - hs, oy - s, text="1", fill=fill, font=font, tags=tags)

    def _draw_pen(self):
        ox, oy = self._calc_true_origin()
        (x, y) = self._pen.pos.copy()
        x = ox + x * self._unit_size
        y = oy - y * self._unit_size
        r = self._pen.width / 2 + 2
        c = [x - r, y - r, x + r, y + r]
        tags = ("pen",)
        fill = self._pen.color if self._pen.can_draw else "gray95"
        outline = "white" if self._pen.can_draw else self._pen.color
        self.create_oval(c, width=2, outline=outline, fill=fill, tags=tags)

    def _draw_lines(self):
        for line in self._lines:
            self._draw_line(line)

    def _draw_line(self, line):
        ox, oy = self._calc_true_origin()
        coords = line[0].copy()
        width = line[1]
        color = line[2]
        coords[0] = ox + coords[0] * self._unit_size  # xs
        coords[1] = oy - coords[1] * self._unit_size  # ys
        coords[2] = ox + coords[2] * self._unit_size  # xs
        coords[3] = oy - coords[3] * self._unit_size  # ys
        self.create_line(coords, fill=color, width=width, tags=("lines",))

    def _draw_grid(self):
        self._draw_verticals()
        self._draw_horizontals()
        self._draw_x_arrow()
        self._draw_y_arrow()

    def _draw_x_arrow(self):
        _, y = self._calc_true_origin()
        x = self._width - 10
        arrowhead_x = x - self._arrow_size
        arrowhead_top = y - self._arrow_size / 2
        arrowhead_bottom = y + self._arrow_size / 2
        coords = [
            # shaft
            10,
            y,
            x,
            y,
            # left side
            arrowhead_x,
            arrowhead_top,
            # central
            x,
            y,
            # right side
            arrowhead_x,
            arrowhead_bottom,
        ]
        self.create_line(coords, fill="gray40", width=2, tags=("grid",))

    def _draw_y_arrow(self):
        x, _ = self._calc_true_origin()
        y = 10
        arrowhead_y = y + self._arrow_size
        arrowhead_left = x - self._arrow_size / 2
        arrowhead_right = x + self._arrow_size / 2
        coords = [
            # shaft
            x,
            self._height - 10,
            x,
            y,
            # left side
            arrowhead_left,
            arrowhead_y,
            # central
            x,
            y,
            # right side
            arrowhead_right,
            arrowhead_y,
        ]
        self.create_line(coords, fill="gray40", width=2, tags=("grid",))

    def _draw_verticals(self):
        line_count = int(self._width / self._unit_size) + 1
        x, _ = self._calc_true_origin()
        delta = x % self._unit_size
        for i in range(line_count):
            x = i * self._unit_size + delta
            coords = [x, 0, x, self._height]
            self.create_line(coords, fill="gray70", width=1, tags=("grid",))

    def _draw_horizontals(self):
        line_count = int(self._height / self._unit_size) + 1
        _, y = self._calc_true_origin()
        delta = y % self._unit_size
        for i in range(line_count):
            y = i * self._unit_size + delta
            coords = [0, y, self._width, y]
            self.create_line(coords, fill="gray70", width=1, tags=("grid",))



"""
This file is part of eduworld package.

This is a main application window class with Canvas and board and etc

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import sys

from ..application import Application


class RobotApplication(Application):
    """Defines an Tkinter application with canvas, board and other stuff"""

    def __init__(self, title=""):
        super().__init__(title)
        self.initialized = False
        self._board = None

    def initialize_app(self):
        super().initialize_app()
        if self._canvas.board is None or not self._canvas.board.initialized:
            print("Board is not set or not initialized! Exiting!")
            sys.exit(1)



"""
This file is part of eduworld package.

This is a main application window class with Canvas and board and etc

=== LICENSE INFO ===

Copyright (c) 2024 - Stanislav Grinkov

The eduworld package is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

The package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the algoworld package.
If not, see `<https://www.gnu.org/licenses/>`_.
"""

import tkinter as tk
import time
import sys
from concurrent.futures import ThreadPoolExecutor


class Application:
    """Defines an Tkinter application with canvas, board and other stuff"""

    def __init__(self, title=""):
        background = "gray75"
        window_width = 800
        window_height = 600
        self._canvas = None
        self.root: tk.Tk = tk.Tk()
        self.root.title("AlgoWorld" if title == "" else f"AlgoWorld-{title}")
        self.root.configure(background=background)
        self.root.rowconfigure(0, weight=1)
        self.root.columnconfigure(0, weight=1)
        self.root.geometry(f"{window_width}x{window_height}")
        self.root.bind("<Q>", lambda _: self._quit())
        self.root.bind("<q>", lambda _: self._quit())

    def set_canvas(self, canvas):
        """Set application canvas"""
        self._canvas = canvas

    def _quit(self):
        self.root.quit()
        self.root.quit()

    def initialize_app(self) -> bool:
        """Extend this and do required checks to make sure
        that your app is properly initialized"""
        if self._canvas is None:
            print("Canvas not set! Please call app.set_canvas()")
            sys.exit(1)

    def run(self) -> None:
        """Runs tkinter TK main loop in a separate thread, to allow
        this main thread to continue the work.
        You have ot call app.shutdown() function to keep the window open
        when your code done its work"""
        self.initialize_app()
        with ThreadPoolExecutor(max_workers=1) as ex:
            ex.submit(self.root.mainloop)
        self._canvas.redraw(True)
        self.root.update()
        time.sleep(1)

    def shutdown(self, keep_window: bool = False) -> None:
        """Join tkinter Tk mainloop to the main thread to keep the window open"""
        if keep_window:
            self.root.mainloop()
        else:
            time.sleep(3)
